'use strict';

Object.defineProperty(exports, "__esModule", {
  value: true
});
exports.queryASTToSqlAST = queryASTToSqlAST;
exports.getGraphQLType = getGraphQLType;
exports.pruneDuplicateSqlDeps = pruneDuplicateSqlDeps;

var _assert = require('assert');

var _assert2 = _interopRequireDefault(_assert);

var _lodash = require('lodash');

var _aliasNamespace = require('./aliasNamespace');

var _aliasNamespace2 = _interopRequireDefault(_aliasNamespace);

var _util = require('./util');

function _interopRequireDefault(obj) { return obj && obj.__esModule ? obj : { default: obj }; }

function queryASTToSqlAST(resolveInfo, options) {
  const namespace = new _aliasNamespace2.default(options.dialect === 'oracle' ? true : options.minify);

  const sqlAST = {};

  const fieldNodes = resolveInfo.fieldNodes || resolveInfo.fieldASTs;
  _assert2.default.equal(fieldNodes.length, 1, 'We thought this would always have a length of 1. FIX ME!!');

  const queryAST = fieldNodes[0];

  const parentType = resolveInfo.parentType;
  getGraphQLType(queryAST, parentType, sqlAST, resolveInfo.fragments, resolveInfo.variableValues, namespace, 0, options);

  _assert2.default.equal(sqlAST.type, 'table', 'Must call joinMonster in a resolver on a field where the type is decorated with "sqlTable".');

  pruneDuplicateSqlDeps(sqlAST, namespace);

  return sqlAST;
}

function getGraphQLType(queryASTNode, parentTypeNode, sqlASTNode, fragments, variables, namespace, depth, options) {
  const fieldName = queryASTNode.name.value;

  if (fieldName.slice(0, 2) === '__') {
    sqlASTNode.type = 'noop';
    return;
  }

  let field = parentTypeNode._fields[fieldName];
  if (!field) {
    throw new Error(`The field "${fieldName}" is not in the ${parentTypeNode.name} type.`);
  }

  let grabMany = false;

  let gqlType = stripNonNullType(field.type);

  if (queryASTNode.arguments.length) {
    const args = sqlASTNode.args = {};
    for (let arg of queryASTNode.arguments) {
      args[arg.name.value] = parseArgValue(arg.value, variables);
    }
  }

  if (gqlType.constructor.name === 'GraphQLList') {
    gqlType = stripNonNullType(gqlType.ofType);
    grabMany = true;
  }

  if (gqlType.constructor.name === 'GraphQLObjectType' && gqlType._fields.edges && gqlType._fields.pageInfo) {
    grabMany = true;

    const stripped = stripRelayConnection(field, queryASTNode, fragments);

    gqlType = stripped.gqlType;
    queryASTNode = stripped.queryASTNode;

    if (field.sqlPaginate) {
      sqlASTNode.paginate = true;
      getSortColumns(field, sqlASTNode);
    }
  } else {
    if (field.sqlPaginate) {
      throw new Error(`To paginate the ${gqlType.name} type, it must be a GraphQLObjectType that fulfills the relay spec. The type must have a "pageInfo" and "edges" field. https://facebook.github.io/relay/graphql/connections.htm`);
    }
  }

  const config = gqlType._typeConfig;

  if (gqlType.constructor.name === 'GraphQLObjectType' && config.sqlTable) {
    if (depth >= 1) {
      (0, _assert2.default)(field.sqlJoin || field.sqlBatch || field.junctionTable, `If an Object type maps to a SQL table and has a child which is another Object type that also maps to a SQL table, you must define "sqlJoin", "sqlBatch", or "junctionTable" on that field to tell joinMonster how to fetch it. Check the "${fieldName}" field on the "${parentTypeNode.name}" type.`);
    }
    handleTable(sqlASTNode, queryASTNode, field, gqlType, fragments, variables, namespace, grabMany, depth, options);
  } else if (gqlType.constructor.name === 'GraphQLUnionType' && config.sqlTable) {
    const config = gqlType._typeConfig;

    sqlASTNode.type = 'table';
    sqlASTNode.name = config.sqlTable;

    sqlASTNode.as = namespace.generate('table', field.name);

    const children = sqlASTNode.children = [];
    handleUniqueKey(config, children, namespace);
    if (config.typeHint) {
      children.push({
        type: 'column',
        name: config.typeHint,
        fieldName: config.typeHint,
        as: namespace.generate('column', config.typeHint)
      });
    }

    sqlASTNode.fieldName = field.name;
    sqlASTNode.grabMany = grabMany;

    if (field.where) {
      sqlASTNode.where = field.where;
    }
    if (field.sqlJoin) {
      sqlASTNode.sqlJoin = field.sqlJoin;
    }
    if (queryASTNode.selectionSet) {
      handleUnionSelections(children, queryASTNode.selectionSet.selections, gqlType, fragments, variables, namespace, depth, options);
    }
  } else if (field.sqlExpr) {
    sqlASTNode.type = 'expression';
    sqlASTNode.sqlExpr = field.sqlExpr;
    sqlASTNode.fieldName = field.name;
    sqlASTNode.as = namespace.generate('column', field.name);
  } else if (field.sqlColumn || !field.resolve) {
    sqlASTNode.type = 'column';
    sqlASTNode.name = field.sqlColumn || field.name;
    sqlASTNode.fieldName = field.name;
    sqlASTNode.as = namespace.generate('column', sqlASTNode.name);
  } else if (field.sqlDeps) {
    sqlASTNode.type = 'columnDeps';
    sqlASTNode.names = field.sqlDeps;
  } else {
    sqlASTNode.type = 'noop';
  }
}

function handleTable(sqlASTNode, queryASTNode, field, gqlType, fragments, variables, namespace, grabMany, depth, options) {
  const config = gqlType._typeConfig;

  sqlASTNode.type = 'table';
  sqlASTNode.name = config.sqlTable;

  sqlASTNode.as = namespace.generate('table', field.name);

  if (field.orderBy && !sqlASTNode.orderBy) {
    handleOrderBy(sqlASTNode, field);
  }

  const children = sqlASTNode.children = [];

  sqlASTNode.fieldName = field.name;
  sqlASTNode.grabMany = grabMany;

  if (field.where) {
    sqlASTNode.where = field.where;
  }
  if (field.sqlJoin) {
    sqlASTNode.sqlJoin = field.sqlJoin;
  } else if (field.junctionTable || field.joinTable) {
    (0, _assert2.default)(field.sqlJoins || field.junctionBatch, 'Must define `sqlJoins` (plural) or `junctionBatch` for a many-to-many.');
    if (field.joinTable) {
      console.warn('The `joinTable` is deprecated. Rename to `junctionTable`.');
    }
    const junctionTable = field.junctionTable || field.joinTable;
    sqlASTNode.junctionTable = junctionTable;
    sqlASTNode.junctionTableAs = namespace.generate('table', junctionTable.replace(/[^a-zA-Z0-9]/g, '_').slice(1, 10));
    if (field.sqlJoins) {
      sqlASTNode.sqlJoins = field.sqlJoins;
    } else {
      if (typeof field.junctionTableKey === 'string') {
        children.push({
          type: 'column',
          name: field.junctionTableKey,
          fieldName: field.junctionTableKey,
          fromOtherTable: sqlASTNode.junctionTableAs,
          as: namespace.generate('column', field.junctionTableKey)
        });
      } else if (Array.isArray(field.junctionTableKey)) {
        const clumsyName = toClumsyName(field.junctionTableKey);
        children.push({
          type: 'composite',
          name: field.junctionTableKey,
          fieldName: clumsyName,
          fromOtherTable: sqlASTNode.junctionTableAs,
          as: namespace.generate('column', clumsyName)
        });
      }
      sqlASTNode.junctionBatch = {
        sqlJoin: field.junctionBatch.sqlJoin,
        thisKey: {
          type: 'column',
          name: field.junctionBatch.thisKey,
          fieldName: field.junctionBatch.thisKey,
          fromOtherTable: sqlASTNode.junctionTableAs,
          as: namespace.generate('column', field.junctionBatch.thisKey)
        },
        parentKey: {
          type: 'column',
          name: field.junctionBatch.parentKey,
          fieldName: field.junctionBatch.parentKey,
          as: namespace.generate('column', field.junctionBatch.parentKey)
        }
      };
    }
  } else if (field.sqlBatch) {
    sqlASTNode.sqlBatch = {
      thisKey: {
        type: 'column',
        name: field.sqlBatch.thisKey,
        fieldName: field.sqlBatch.thisKey,
        as: namespace.generate('column', field.sqlBatch.thisKey)
      },
      parentKey: {
        type: 'column',
        name: field.sqlBatch.parentKey,
        fieldName: field.sqlBatch.parentKey,
        as: namespace.generate('column', field.sqlBatch.parentKey)
      }
    };
  }

  handleUniqueKey(config, children, namespace);

  if (sqlASTNode.paginate) {
    handleColumnsRequiredForPagination(sqlASTNode, namespace);
  }

  if (queryASTNode.selectionSet) {
    handleSelections(children, queryASTNode.selectionSet.selections, gqlType, fragments, variables, namespace, depth, options);
  }
}

function handleUnionSelections(children, selections, gqlType, fragments, variables, namespace, depth, options) {
  for (let selection of selections) {
    switch (selection.kind) {
      case 'Field':
        const newNode = {};
        children.push(newNode);
        getGraphQLType(selection, gqlType, newNode, fragments, variables, namespace, depth + 1, options);
        break;

      case 'InlineFragment':
        {
          const selectionNameOfType = selection.typeCondition.name.value;

          const deferToType = gqlType._types.find(type => type.name === selectionNameOfType);
          handleSelections(children, selection.selectionSet.selections, deferToType, fragments, variables, namespace, depth, options);
        }
        break;

      case 'FragmentSpread':
        {
          const fragmentName = selection.name.value;
          const fragment = fragments[fragmentName];

          const fragmentNameOfType = fragment.typeCondition.name.value;

          const deferToType = gqlType._types.find(type => type.name === fragmentNameOfType);
          handleSelections(children, fragment.selectionSet.selections, deferToType, fragments, variables, namespace, depth, options);
        }
        break;
      default:
        throw new Error('Unknown selection kind: ' + selection.kind);
    }
  }
}

function handleSelections(children, selections, gqlType, fragments, variables, namespace, depth, options) {
  for (let selection of selections) {
    switch (selection.kind) {
      case 'Field':
        const newNode = {};
        children.push(newNode);
        getGraphQLType(selection, gqlType, newNode, fragments, variables, namespace, depth + 1, options);
        break;

      case 'InlineFragment':
        {
          const selectionNameOfType = selection.typeCondition.name.value;
          const sameType = selectionNameOfType === gqlType.name;
          const interfaceType = (gqlType._interfaces || []).map(iface => iface.name).includes(selectionNameOfType);
          if (sameType || interfaceType) {
            handleSelections(children, selection.selectionSet.selections, gqlType, fragments, variables, namespace, depth, options);
          }
        }
        break;

      case 'FragmentSpread':
        {
          const fragmentName = selection.name.value;
          const fragment = fragments[fragmentName];

          const fragmentNameOfType = fragment.typeCondition.name.value;
          const sameType = fragmentNameOfType === gqlType.name;
          const interfaceType = gqlType._interfaces.map(iface => iface.name).indexOf(fragmentNameOfType) >= 0;
          if (sameType || interfaceType) {
            handleSelections(children, fragment.selectionSet.selections, gqlType, fragments, variables, namespace, depth, options);
          }
        }
        break;
      default:
        throw new Error('Unknown selection kind: ' + selection.kind);
    }
  }
}

function handleUniqueKey(config, children, namespace) {
  if (!config.uniqueKey) {
    throw new Error(`You must specify the "uniqueKey" on the GraphQLObjectType definition of ${config.sqlTable}`);
  }
  if (typeof config.uniqueKey === 'string') {
    children.push({
      type: 'column',
      name: config.uniqueKey,
      fieldName: config.uniqueKey,
      as: namespace.generate('column', config.uniqueKey)
    });
  } else if (Array.isArray(config.uniqueKey)) {
    const clumsyName = toClumsyName(config.uniqueKey);
    children.push({
      type: 'composite',
      name: config.uniqueKey,
      fieldName: clumsyName,
      as: namespace.generate('column', clumsyName)
    });
  }
}

function toClumsyName(keyArr) {
  return keyArr.map(name => name.slice(0, 3)).join('#');
}

function handleColumnsRequiredForPagination(sqlASTNode, namespace) {
  if (sqlASTNode.sortKey) {
    for (let column of (0, _util.wrap)(sqlASTNode.sortKey.key)) {
      const newChild = {
        type: 'column',
        name: column,
        fieldName: column,
        as: namespace.generate('column', column)
      };

      if (sqlASTNode.junctionTable) {
        newChild.fromOtherTable = sqlASTNode.junctionTableAs;
      }
      sqlASTNode.children.push(newChild);
    }
  } else if (sqlASTNode.orderBy) {
    const newChild = {
      type: 'column',
      name: '$total',
      fieldName: '$total',
      as: namespace.generate('column', '$total')
    };
    if (sqlASTNode.junctionTable) {
      newChild.fromOtherTable = sqlASTNode.junctionTableAs;
    }
    sqlASTNode.children.push(newChild);
  }
}

function stripRelayConnection(field, queryASTNode, fragments) {
  const gqlType = field.type._fields.edges.type.ofType._fields.node.type;

  const args = queryASTNode.arguments;

  const edges = spreadFragments(queryASTNode.selectionSet.selections, fragments, field.type.name).find(selection => selection.name.value === 'edges');
  if (edges) {
    queryASTNode = spreadFragments(edges.selectionSet.selections, fragments, field.type.name).find(selection => selection.name.value === 'node') || {};
  } else {
    queryASTNode = {};
  }

  queryASTNode.arguments = args;
  return { gqlType, queryASTNode };
}

function stripNonNullType(type) {
  return type.constructor.name === 'GraphQLNonNull' ? type.ofType : type;
}

function pruneDuplicateSqlDeps(sqlAST, namespace) {
  const deps = new Set();
  const children = sqlAST.children || [];

  for (let i = children.length - 1; i >= 0; i--) {
    const child = children[i];
    if (child.type === 'columnDeps') {
      child.names.forEach(name => deps.add(name));
      children.splice(i, 1);
    } else if (child.type === 'table') {
      pruneDuplicateSqlDeps(child, namespace);
    }
  }

  const newNode = {
    type: 'columnDeps',
    names: {}
  };
  deps.forEach(name => {
    newNode.names[name] = namespace.generate('column', name);
  });
  children.push(newNode);
}

function parseArgValue(value, variableValues) {
  if (value.kind === 'Variable') {
    const variableName = value.name.value;
    return variableValues[variableName];
  }

  switch (value.kind) {
    case 'IntValue':
      return parseInt(value.value);
    case 'FloatValue':
      return parseFloat(value.value);
    case 'ListValue':
      return value.values.map(value => parseArgValue(value, variableValues));
    default:
      return value.value;
  }
}

function getSortColumns(field, sqlASTNode) {
  if (field.sortKey) {
    if (typeof field.sortKey === 'function') {
      sqlASTNode.sortKey = field.sortKey(sqlASTNode.args);
    } else {
      sqlASTNode.sortKey = field.sortKey;
    }
  } else if (field.orderBy) {
    handleOrderBy(sqlASTNode, field);
  } else {
    throw new Error('"sortKey" or "orderBy" required if "sqlPaginate" is true');
  }
}

function handleOrderBy(sqlASTNode, field) {
  if (typeof field.orderBy === 'function') {
    sqlASTNode.orderBy = field.orderBy(sqlASTNode.args || {});
  } else {
    sqlASTNode.orderBy = field.orderBy;
  }
}

function spreadFragments(selections, fragments, typeName) {
  return (0, _lodash.flatMap)(selections, selection => {
    switch (selection.kind) {
      case 'FragmentSpread':
        const fragmentName = selection.name.value;
        const fragment = fragments[fragmentName];
        return spreadFragments(fragment.selectionSet.selections, fragments, typeName);
      case 'InlineFragment':
        if (selection.typeCondition.name.value === typeName) {
          return spreadFragments(selection.selectionSet.selections, fragments, typeName);
        } else {
          return [];
        }
      default:
        return selection;
    }
  });
}